#!/usr/bin/env python

from Bio.Seq import translate

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description='Translate FASTA and write the result to standard output.')

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()

    for read in parseFASTACommandLineOptions(args):
        print(f'>{read.id}\n{translate(read.sequence)}')



from unittest import TestCase

from dark.codonDistance import codonInformation, findDistance


class TestCodonInformation(TestCase):
    """
    Tests for the codonInformation function.
    """
    def testReturnRightCodon(self):
        """
        The function must return the right codon distances.
        """
        result = codonInformation(['GAT', 'GAC'], ['GCC', 'GCA'])
        self.assertEqual([('GAC', 'GCC')], result[1])
        self.assertEqual([('GAT', 'GCC'), ('GAT', 'GCA'), ('GAC', 'GCA')],
                         result[2])

    def testReturnRightCodonWithTransitionCount(self):
        """
        The function must return the right codon distances and the correct
        number of transitions.
        """
        result = codonInformation(['GAT', 'GAC'], ['GCC', 'AGT'],
                                  countTransitions=True)
        self.assertEqual([('GAC', 'GCC', 0)], result[1])
        self.assertEqual([('GAT', 'GCC', 1), ('GAT', 'AGT', 2)], result[2])
        self.assertEqual([('GAC', 'AGT', 3)], result[3])


class TestFindDistance(TestCase):
    """
    Tests for the findDistance function.
    """
    def testFindDistanceThree(self):
        """
        findDistance must return the right distance if codons are different.
        """
        distance = findDistance('ACA', 'TGT')
        self.assertEqual(3, distance)

    def testFindDistanceZero(self):
        """
        findDistance must return the right distance if codons are identical.
        """
        distance = findDistance('ACA', 'ACA')
        self.assertEqual(0, distance)



#!/usr/bin/env python

import sys
import argparse
import re

from dark.reads import (
    addFASTACommandLineOptions, parseFASTACommandLineOptions, Reads)

parser = argparse.ArgumentParser(
    description=('Write sorted FASTA/Q to stdout. Sorting is by sequence id, '
                 'then by sequence, then by quality (if FASTQ)'))

# TODO: Support multiple regexes and apply them one by one to allow matching on
# different regions in an arbitrary order.
parser.add_argument(
    '--regex',
    help=('A regular expression to specify a region (or regions) to extract '
          'from sequence ids to sort on. Regions will be converted to '
          'integers if possible.'))


parser.add_argument(
    '--reverse', '-r', action='store_true',
    help='Sort by decreasing value.')

addFASTACommandLineOptions(parser)
args = parser.parse_args()
reads = parseFASTACommandLineOptions(args)

if args.regex:
    regex = re.compile(args.regex)

    if regex.groups == 0:
        print('You have passed a regular expression that has no capturing '
              'group(s) specified using (...).', file=sys.stderr)
        sys.exit(1)

    def key(read):
        result = []
        match = regex.search(read.id)
        if match:
            for text in match.groups():
                try:
                    value = int(text)
                except ValueError:
                    value = text
                result.append(value)
        return result
else:
    def key(read):
        return read.id


Reads(sorted(reads, key=key, reverse=args.reverse)).save(
    sys.stdout, 'fastq' if args.fastq else 'fasta')



#!/usr/bin/env python

import sys
from collections import defaultdict
import csv
from operator import itemgetter

from dark.dna import AMBIGUOUS
from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

# These nucleotide codes will be listed first if --sortChars is used.
SPECIAL_CHARS = 'ACGT-?N'


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description='Print coverage information to standard output.')

    parser.add_argument(
        '--digits', default=4, type=int,
        help=('The number of digits of precision to use for floating point '
              'outputs (coverage, gc fraction).'))

    parser.add_argument(
        '--noNames', dest='printNames', action='store_false',
        help='Do not print the ids of input sequences.')

    parser.add_argument(
        '--strict', action='store_true',
        help=('Treat all ambiguous nucleotides as not covered. Additional '
              'characters specified with --chars will also be counted as not '
              'covered.'))

    parser.add_argument(
        '--gc', action='store_true',
        help='Include the GC percentage in the output.')

    parser.add_argument(
        '--reverse', '-r', action='store_true',
        help='Sort lines by decreasing value.')

    parser.add_argument(
        '--sortBy',
        help=('What to sort lines by. Either one of id, coverage, length, gc, '
              'not-covered or else a single nucleotide character.'))

    parser.add_argument(
        '--header', action='store_true',
        help='Write a header line and in cells print just the values.')

    parser.add_argument(
        '--sep', default='\t',
        help='The output separator character.')

    parser.add_argument(
        '--printCounts', action='store_true',
        help='Print the count for each character found in each read.')

    parser.add_argument(
        '--sortChars', action='store_true',
        help='Sort the output columns.')

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)
    coveredChars = set('AGCT') if args.strict else set(AMBIGUOUS)
    data = []
    sortBy = args.sortBy
    keepCounts = args.printCounts or args.gc or (sortBy and len(sortBy) == 1)

    for read in reads:
        notCoveredCount = 0
        if keepCounts:
            counts = defaultdict(int)
        for char in read.sequence.upper():
            notCoveredCount += (char not in coveredChars)
            if keepCounts:
                counts[char] += 1

        length = len(read)

        # Sanity check.
        if keepCounts:
            assert sum(counts.values()) == length

        coverage = round(
            (length - notCoveredCount) / length if length else 0.0,
            args.digits)

        readData = {
            'id': read.id,
            'length': len(read),
            'not-covered': notCoveredCount,
            'coverage': coverage,
        }

        if args.printCounts:
            readData['counts'] = counts

        if args.gc:
            gc = 0
            for char in 'GC':
                if char in counts:
                    gc += counts[char]
            readData['gc'] = round(gc / length, args.digits)

        data.append(readData)

    if not data:
        sys.exit(0)

    writerow = csv.writer(sys.stdout, delimiter=args.sep).writerow

    if args.sortChars:
        allChars = set()
        for d in data:
            allChars.update(d['counts'])

        specials = [char for char in SPECIAL_CHARS if char in allChars]
        allChars = specials + sorted(allChars - set(specials))

    header = args.header

    if header:
        line = ['Id'] if args.printNames else []
        line.extend(('Length', 'Not covered', 'Coverage'))
        if args.gc:
            line.append('GC')
        if args.printCounts and args.sortChars:
            line.extend(allChars)
        writerow(line)

    if sortBy:
        if sortBy in data[0]:
            key = itemgetter(sortBy)
        else:
            if len(sortBy) > 1:
                print('Cannot sort by unknown column {sortBy!r}.',
                      file=sys.stderr)
                sys.exit(1)

            sortBy = sortBy.upper()

            def key(d):
                return d['counts'].get(sortBy, 0)

        data = sorted(data, key=key, reverse=args.reverse)

    for d in data:

        line = [d['id']] if args.printNames else []
        line.append(d['length'] if header else f"length:{d['length']}")
        line.append(d['not-covered'] if header else
                    f"not-covered:{d['not-covered']}")
        line.append(d['coverage'] if header else f"coverage:{d['coverage']}")
        if args.gc:
            line.append(d['gc'] if header else f"gc:{d['gc']}")

        if args.printCounts:
            counts = d['counts']
            if args.sortChars:
                for char in allChars:
                    if char in counts:
                        line.append(
                            counts[char] if header else
                            f"{char}:{counts[char]}")
                    else:
                        line.append('')
            else:
                for char, count in sorted(counts.items()):
                    line.append(count if header else f"{char}:{count}")

        writerow(line)



#!/usr/bin/env python

from __future__ import print_function, division

import argparse
import requests

# See https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch
# for URL format details.

URL = ('https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?'
       'db=%s&id=%s&rettype=%s&retmode=text')

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Fetch a sequence by id from NCBI and write it to stdout '
                 'as FASTA.'))

parser.add_argument(
    'id', help='The id of the sequence to fetch.')

parser.add_argument(
    '--database', default='nucleotide', choices=('nucleotide', 'protein'),
    help='The name of the NCBI database to query.')

parser.add_argument(
    '--format', default='fasta', choices=('fasta', 'gb'),
    help='The format to return.')

args = parser.parse_args()

print(requests.get(
    URL % (args.database, args.id, args.format)
).text.rstrip('\n'))



#!/usr/bin/env python

import sys
import argparse
from json import dump

from dark.sam import DistanceMatrix


def main(args):
    """
    Print a JSON object with reference to read scores.

    @param args: A C{Namespace} object returned by argparse, containing
        argument values passed on the command line.
    """
    dm = DistanceMatrix()
    for filename in args.samFile:
        dm.addFile(filename, scoreTag=args.scoreTag)

    if args.minMatchingReads is None:
        wantedReferenceIds = sorted(set(dm.scores))
    else:
        wantedReferenceIds = sorted(set(
            referenceId for (referenceId, reads) in dm.scores.items()
            if len(reads) >= args.minMatchingReads))
        if args.verbose:
            print(f'Found {len(wantedReferenceIds)} references with at least '
                  f'{args.minMatchingReads} matching reads.', file=sys.stderr)

    if args.verbose:
        for referenceId in wantedReferenceIds:
            nReads = len(dm.scores[referenceId])
            print(f'Reference {referenceId!r} matched {nReads} reads.')

    dump(dict((id_, dm.scores[id_]) for id_ in wantedReferenceIds),
         sys.stdout, sort_keys=True, indent=4)
    print()


def makeParser():
    """
    Make a command-line argument parser.

    @return: An C{argparse.ArgumentParser} instance.
    """
    parser = argparse.ArgumentParser(
        description=('Print a JSON object containing reference to read '
                     'distances extracted from a SAM file.'))

    parser.add_argument(
        '--samFile', action='append', required=True,
        help='The SAM file(s) to load. May be repeated.')

    parser.add_argument(
        '--minMatchingReads', type=int,
        help=('The minimum number of reads that must match a reference for it '
              'to be included.'))

    parser.add_argument(
        '--scoreTag',
        help=('The score tag to use for the alignment score. If not given, '
              '1 will be used to indicate that a read matched a reference '
              '(non-matches are not included). The default is no score tag, '
              'which is not that useful. A good choice is "AS", for the '
              'alignment score, but that has to be present in the SAM file, '
              'which means that the aligner (bowtie2, bwa, etc. has to have '
              'produced such a tag.'))

    parser.add_argument(
        '--verbose', action='store_true',
        help='Print extra information.')

    return parser


if __name__ == '__main__':
    main(makeParser().parse_args())
